/*********************************************************************
 * Software License Agreement (BSD License)
 * 
 *  Copyright (c) 2008, Willow Garage, Inc.
 *  All rights reserved.
 * 
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 * 
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above
 *     copyright notice, this list of conditions and the following
 *     disclaimer in the documentation and/or other materials provided
 *     with the distribution.
 *   * Neither the name of Willow Garage nor the names of its
 *     contributors may be used to endorse or promote products derived
 *     from this software without specific prior written permission.
 * 
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 *  FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 *  BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 *  LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 *  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 *  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 *  ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 *  POSSIBILITY OF SUCH DAMAGE.
 *
 *
 *********************************************************************/

/* \author: Ioan Sucan, Sachin Chitta*/

#include <ros/ros.h>
#include <actionlib/client/simple_action_client.h>
#include <arm_navigation_msgs/MoveArmAction.h>

#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <ros/ros.h>
#include <gtest/gtest.h>

#include <arm_navigation_tests/arm_navigation_utils.h>

void spinThread()
{
  ros::spin();
}

TEST(MoveArm, goToJointGoal)
{
  ros::NodeHandle nh;
  ros::NodeHandle private_handle("~");
  actionlib::SimpleActionClient<arm_navigation_msgs::MoveArmAction> move_arm(nh, "move_right_arm");
  boost::thread spin_thread(&spinThread);

  move_arm.waitForServer();
  ROS_INFO("Connected to server");

  ros::WallDuration(1.0).sleep();
      
  arm_navigation_msgs::MoveArmGoal goalB;
  
  arm_navigation_msgs::CollisionObject box;
  box.operation.operation = arm_navigation_msgs::CollisionObjectOperation::ADD;
  box.header.frame_id = "r_gripper_r_finger_tip_link";
  box.header.stamp = ros::Time::now();
  arm_navigation_msgs::Shape object;
  object.type = arm_navigation_msgs::Shape::BOX;
  object.dimensions.resize(3);
  object.dimensions[0] = .05;
  object.dimensions[1] = .05;
  object.dimensions[2] = .05;
  geometry_msgs::Pose pose;
  pose.position.x = 0.0;
  pose.position.y = 0.0;
  pose.position.z = 0.0;
  pose.orientation.x = 0;
  pose.orientation.y = 0;
  pose.orientation.z = 0;
  pose.orientation.w = 1;
  box.shapes.push_back(object);
  box.poses.push_back(pose);
  box.id = "box";
  goalB.planning_scene_diff.collision_objects.push_back(box);

  std::vector<std::string> names(7);
  names[0] = "r_shoulder_pan_joint";
  names[1] = "r_shoulder_lift_joint";
  names[2] = "r_upper_arm_roll_joint";
  names[3] = "r_elbow_flex_joint";
  names[4] = "r_forearm_roll_joint";
  names[5] = "r_wrist_flex_joint";
  names[6] = "r_wrist_roll_joint";

  goalB.motion_plan_request.group_name = "right_arm";
  goalB.motion_plan_request.num_planning_attempts = 1;
  goalB.motion_plan_request.allowed_planning_time = ros::Duration(5.0);

  private_handle.param<std::string>("planner_id",goalB.motion_plan_request.planner_id,std::string(""));
  private_handle.param<std::string>("planner_service_name",goalB.planner_service_name,std::string("/ompl_planning/plan_kinematic_path"));
    
  goalB.motion_plan_request.goal_constraints.joint_constraints.resize(names.size());
  for (unsigned int i = 0 ; i < goalB.motion_plan_request.goal_constraints.joint_constraints.size(); ++i)
  {
    // goalB.motion_plan_request.goal_constraints.joint_constraints[i].header.stamp = ros::Time::now();
    // goalB.motion_plan_request.goal_constraints.joint_constraints[i].header.frame_id = "base_link";
    goalB.motion_plan_request.goal_constraints.joint_constraints[i].joint_name = names[i];
    goalB.motion_plan_request.goal_constraints.joint_constraints[i].position = 0.0;
    goalB.motion_plan_request.goal_constraints.joint_constraints[i].tolerance_below = 0.1;
    goalB.motion_plan_request.goal_constraints.joint_constraints[i].tolerance_above = 0.1;
  }
    
  goalB.motion_plan_request.goal_constraints.joint_constraints[0].position = -2.0;
  goalB.motion_plan_request.goal_constraints.joint_constraints[3].position = -0.2;
  goalB.motion_plan_request.goal_constraints.joint_constraints[5].position = -0.20;
   
  bool success = false;
  
  bool finished_within_time = false;
  move_arm.sendGoal(goalB);
  finished_within_time = move_arm.waitForResult(ros::Duration(200.0));
  actionlib::SimpleClientGoalState state = move_arm.getState();
  success = (state == actionlib::SimpleClientGoalState::SUCCEEDED);

  EXPECT_FALSE(success);
  arm_navigation_msgs::MoveArmResultConstPtr resb = move_arm.getResult();

  ASSERT_TRUE(resb->contacts.size() > 0);
  
  for(unsigned int i = 0; i < resb->contacts.size(); i++) {
    arm_navigation_msgs::AllowedContactSpecification acs;
    acs.shape.type = acs.shape.SPHERE;
    acs.shape.dimensions.push_back(.1);
    acs.pose_stamped.header = resb->contacts[i].header;
    acs.pose_stamped.pose.position = resb->contacts[i].position;
    acs.pose_stamped.pose.orientation.w = 1.0;
    acs.link_names.push_back(resb->contacts[i].contact_body_1);
    acs.link_names.push_back(resb->contacts[i].contact_body_2);
    acs.penetration_depth = 100;
    goalB.planning_scene_diff.allowed_contacts.push_back(acs);
  }
  ROS_INFO_STREAM("Setting " << goalB.planning_scene_diff.allowed_contacts.size() << " allowed contacts");
  move_arm.sendGoal(goalB);
  finished_within_time = move_arm.waitForResult(ros::Duration(200.0));
  state = move_arm.getState();
  success = (state == actionlib::SimpleClientGoalState::SUCCEEDED);

  ASSERT_TRUE(success);
  ros::shutdown();
  spin_thread.join();
}

int main(int argc, char **argv){
  testing::InitGoogleTest(&argc, argv);
  ros::init (argc, argv, "move_arm_regression_test");
  return RUN_ALL_TESTS();
}
