/*
 * Copyright (c) 2011, Willow Garage, Inc.
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the <ORGANIZATION> nor the names of its
 *       contributors may be used to endorse or promote products derived from
 *       this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

// Author: E. Gil Jones

#include <termios.h>
#include <signal.h>
#include <cstdio>
#include <cstdlib>
#include <unistd.h>
#include <math.h>

#include <ros/ros.h>

#include <arm_navigation_msgs/PlanningScene.h>
#include <planning_environment/models/model_utils.h>
#include <trajectory_msgs/JointTrajectory.h>
#include <move_arm_warehouse/move_arm_warehouse_logger_reader.h>
#include <collision_proximity/collision_proximity_space.h>
#include <arm_navigation_msgs/JointLimits.h>
#include <spline_smoother/cubic_trajectory.h>

struct TrajectoryMetrics {

  TrajectoryMetrics() :
    average_closest_collision_distance_(DBL_MAX),
    overall_closest_collision_distance_(DBL_MAX),
    total_path_length_(0.0),
    spline_time_(0.0),
    processing_time_(0.0)
  {
  }

  double average_closest_collision_distance_;
  double overall_closest_collision_distance_;
  double total_path_length_;
  double spline_time_;
  ros::Duration processing_time_;
};

struct MoveArmInstanceMetrics {
  std::map<std::string, TrajectoryMetrics> stage_metrics_;
  arm_navigation_msgs::ArmNavigationErrorCodes error_code_;
  std::string pipeline_stage_;
  bool has_path_constraints_;
};

struct TrajectorySourceMetrics {

  TrajectorySourceMetrics() :
    total_average_closest_collision_distance_(0.0), 
    total_overall_closest_collision_distance_(0.0), 
    total_total_path_length_(0.0),
    total_spline_time_(0.0),
    total_processing_time_(0.0)
  {
  }

  void addTrajectoryMetric(const TrajectoryMetrics& metric) {
    instance_metrics_.push_back(metric);
    total_spline_time_ += metric.spline_time_;
    total_average_closest_collision_distance_ += metric.average_closest_collision_distance_;
    total_overall_closest_collision_distance_ += metric.overall_closest_collision_distance_;
    total_total_path_length_ += metric.total_path_length_;
    total_processing_time_ += metric.processing_time_;
  }

  double getSizeAsDouble() const {
    return (instance_metrics_.size()*1.0);
  }

  std::vector<TrajectoryMetrics> instance_metrics_;
  std::vector<bool> has_path_constraints;

  double total_average_closest_collision_distance_;
  double total_overall_closest_collision_distance_;
  double total_total_path_length_;
  double total_spline_time_;
  ros::Duration total_processing_time_;
};

struct OverallMetrics {

  OverallMetrics() : 
    successes_(0),
    recoverable_errors_(0),
    invocations_(0)
  {}

  unsigned int successes_;
  unsigned int recoverable_errors_;
  unsigned int invocations_;
  
  std::map<std::string, unsigned int> failure_stage_;

  std::vector<MoveArmInstanceMetrics> move_arm_metric_vector_;
  std::map<std::string, TrajectorySourceMetrics> trajectory_source_metrics_;
};

class WarehouseMetricsEvaluator {

public:

  WarehouseMetricsEvaluator() {
    move_arm_warehouse_reader_ = new move_arm_warehouse::MoveArmWarehouseLoggerReader();

    std::string robot_description_name = root_nh_.resolveName("robot_description", true);

    cps_ = new collision_proximity::CollisionProximitySpace(robot_description_name);

    vis_marker_publisher_ = root_nh_.advertise<visualization_msgs::Marker>("metrics_markers", 128);
    vis_marker_array_publisher_ = root_nh_.advertise<visualization_msgs::MarkerArray>("metrics_markers_array", 128);
  };

  ~WarehouseMetricsEvaluator()
  {
    delete move_arm_warehouse_reader_;
    delete cps_;
  }

  void getJointLimits(const std::vector<std::string>& joint_names, std::vector<arm_navigation_msgs::JointLimits>& limit_vec) 
  {
    limit_vec.clear();
    for(unsigned int i = 0; i < joint_names.size(); i++) {
      if(joint_limits_map_.find(joint_names[i]) == joint_limits_map_.end()) {
        arm_navigation_msgs::JointLimits limits;
        root_nh_.param("joint_limits/"+joint_names[i]+"/min_position", limits.min_position, -std::numeric_limits<double>::max());
        root_nh_.param("joint_limits/"+joint_names[i]+"/max_position", limits.max_position, std::numeric_limits<double>::max());
        root_nh_.param("joint_limits/"+joint_names[i]+"/max_velocity", limits.max_velocity, std::numeric_limits<double>::max());
        root_nh_.param("joint_limits/"+joint_names[i]+"/max_acceleration", limits.max_acceleration, std::numeric_limits<double>::max());
        bool boolean;
        root_nh_.param("joint_limits/"+joint_names[i]+"/has_position_limits", boolean, false);
        limits.has_position_limits = boolean?1:0;
        root_nh_.param("joint_limits/"+joint_names[i]+"/has_velocity_limits", boolean, false);
        limits.has_velocity_limits = boolean?1:0;
        root_nh_.param("joint_limits/"+joint_names[i]+"/has_acceleration_limits", boolean, false);
        limits.has_acceleration_limits = boolean?1:0;
        joint_limits_map_[joint_names[i]] = limits;
        limit_vec.push_back(limits);
      } else {
        limit_vec.push_back(joint_limits_map_[joint_names[i]]);
      }
    }
  }

  TrajectoryMetrics calculateTrajectoryMetrics(const arm_navigation_msgs::RobotState& start_state,
                                               const trajectory_msgs::JointTrajectory& joint_trajectory) {
    TrajectoryMetrics traj_metrics;
    
    std::vector<arm_navigation_msgs::JointLimits> limit_vec;
    getJointLimits(joint_trajectory.joint_names, limit_vec);
    
    trajectory_msgs::JointTrajectory jtraj = joint_trajectory;

    if(joint_trajectory.points.size() == 0) {
      return traj_metrics;
    }

    for(unsigned int i = 0; i < jtraj.points.size(); i++) {
      if(jtraj.points[i].velocities.size() != jtraj.points[i].positions.size()) {
        jtraj.points[i].velocities.resize(jtraj.points[i].positions.size(), 0.0);
      }
    }    
    //create a spline from the trajectory
    spline_smoother::CubicTrajectory trajectory_solver;
    spline_smoother::SplineTrajectory spline;
    
    trajectory_solver.parameterize(jtraj,limit_vec,spline);  
    
    double smoother_time;
    spline_smoother::getTotalTime(spline, smoother_time);   

    traj_metrics.spline_time_ = smoother_time;
    
    jtraj.points.clear();

    unsigned int num_points = 200;

    double t = 0.0;
    std::vector<double> times(num_points);
    for(unsigned int i = 0; i < num_points; i++,t += smoother_time/(1.0*(num_points))) {
      times[i] = t;
    }
    
    if(!spline_smoother::sampleSplineTrajectory(spline, times, jtraj)) {
      ROS_INFO("Spline smoother reports problem");
    }

    //double planner_time = traj.points.back().time_from_start.toSec();
    double total_collision_distance = 0.0;  
    std::string overall_closest_link;
    unsigned int closest_collision_point = 0;
    t = 0.0;
    for(unsigned int i = 0; i < jtraj.points.size(); i++, t += smoother_time/(1.0*(num_points))) {
      jtraj.points[i].time_from_start = ros::Duration(t);
    }

    std::vector<std::string> objs = current_link_names_;
    objs.insert(objs.end(), current_attached_body_names_.begin(), current_attached_body_names_.end());

    std::map<std::string, double> joint_positions;
    for(unsigned int i = 0; i < jtraj.points.size(); i++) {
      for(unsigned int j = 0; j < joint_trajectory.joint_names.size(); j++) {
        joint_positions[joint_trajectory.joint_names[j]] = jtraj.points[i].positions[j];
        if(i != jtraj.points.size()-1) {
          traj_metrics.total_path_length_ += fabs(jtraj.points[i+1].positions[j]-joint_positions[joint_trajectory.joint_names[j]]);
        }
      }
      cps_->getCollisionModelsInterface()->getPlanningSceneState()->setKinematicState(joint_positions);
      cps_->setCurrentGroupState(*cps_->getCollisionModelsInterface()->getPlanningSceneState());

      std::vector<collision_proximity::GradientInfo> gradients;
      cps_->getEnvironmentProximityGradients(gradients, true);

      if(objs.size() != gradients.size()) {
        ROS_WARN_STREAM("Objs size not equal to gradients " << objs.size() << " " << current_link_names_.size() << " " << gradients.size());
        return traj_metrics;
      }
      double this_distance = DBL_MAX;
      std::string this_link;
      for(unsigned int j = 0; j < objs.size(); j++) {
        if(gradients[j].closest_distance < this_distance) {
          if(gradients[j].closest_distance <= 0.0) {
            //ROS_INFO_STREAM("Link " << objs[j] << " apparently in collision " << gradients[j].closest_distance);
          }
          this_distance = gradients[j].closest_distance;
          this_link = objs[j];
        }
      }
      total_collision_distance += this_distance;
      if(this_distance <= traj_metrics.overall_closest_collision_distance_) {
        traj_metrics.overall_closest_collision_distance_ = this_distance;
        overall_closest_link = this_link;
        closest_collision_point = i;
      }
    }

    traj_metrics.average_closest_collision_distance_ = total_collision_distance/jtraj.points.size();

    // for(unsigned int j = 0; j < joint_trajectory.joint_names.size(); j++) {
    //   joint_positions[joint_trajectory.joint_names[j]] = jtraj.points[closest_collision_point].positions[j];
    // }

    // cps_->getCollisionModelsInterface()->getPlanningSceneState()->setKinematicState(joint_positions);
    // cps_->setCurrentGroupState(*cps_->getCollisionModelsInterface()->getPlanningSceneState());

    // std::vector<std::string> link_names;
    // std::vector<std::string> attached_body_names;
    // std::vector<collision_proximity::GradientInfo> gradients;
    // cps_->getStateGradients(link_names, attached_body_names, gradients);

    // visualization_msgs::MarkerArray arr;
    // cps_->getProximityGradientMarkers(link_names,
    //                                   attached_body_names,
    //                                   gradients,
    //                                   "",
    //                                   arr);
    // vis_marker_array_publisher_.publish(arr);


    ROS_INFO_STREAM("Trajectory has closest overall distance " 
                    << traj_metrics.overall_closest_collision_distance_);
    ROS_INFO_STREAM("Path length " << traj_metrics.total_path_length_ << " spline time " << traj_metrics.spline_time_);    
    ROS_INFO_STREAM("Closest link is " << overall_closest_link);
    return traj_metrics;
  }


  void calculateMetrics() {

    std::vector<ros::Time> planning_scene_creation_times;
    move_arm_warehouse_reader_->getAvailablePlanningSceneList("", planning_scene_creation_times);
    
    overall_metrics_.move_arm_metric_vector_.resize(planning_scene_creation_times.size());

    for(unsigned int i = 0; i < planning_scene_creation_times.size(); i++) {
      arm_navigation_msgs::PlanningScene planning_scene;
      if(!move_arm_warehouse_reader_->getPlanningScene("", planning_scene_creation_times[i],
                                                       planning_scene)) {
        ROS_WARN_STREAM("Problem with planning scene for creation time " << planning_scene_creation_times[i]);
        continue;
      }

      if(!cps_->setPlanningScene(planning_scene)) {
        ROS_WARN_STREAM("Bad planning scene " << planning_scene_creation_times[i]);
        continue;
      }
      
      std::vector<std::string> pipeline_stages;
      std::vector<arm_navigation_msgs::ArmNavigationErrorCodes> error_codes;
      std::vector<std::string> trajectories;
      if(!move_arm_warehouse_reader_->getAssociatedOutcomes("", planning_scene_creation_times[i],
                                                            pipeline_stages,
                                                            error_codes,
                                                            trajectories)) {
        ROS_WARN_STREAM("Problem with outcome for creation time " << planning_scene_creation_times[i]);
        arm_navigation_msgs::ArmNavigationErrorCodes error_code;
        error_code.val = 0;
        error_codes.push_back(error_code);
        pipeline_stages.push_back("Unknown");
      }
      overall_metrics_.move_arm_metric_vector_[i].error_code_ = error_codes.back();
      overall_metrics_.move_arm_metric_vector_[i].pipeline_stage_ = pipeline_stages.back();

      overall_metrics_.invocations_++;

      if(error_codes.back().val != error_codes.back().SUCCESS) {
        overall_metrics_.failure_stage_[pipeline_stages.back()]++;
      } else if(error_codes.size() >1) {
        overall_metrics_.recoverable_errors_++;
        overall_metrics_.successes_++;
      } else {
        overall_metrics_.successes_++;
      }

      std::vector<std::string> motion_plan_request_stage_names;
      if(!move_arm_warehouse_reader_->getAssociatedMotionPlanRequestsStageNames("", planning_scene_creation_times[i],
                                                                                motion_plan_request_stage_names)) {
        ROS_WARN_STREAM("No motion plan requests for creation time " << planning_scene_creation_times[i]);
        continue;
      }

      arm_navigation_msgs::MotionPlanRequest motion_plan_request;
      std::string ID;
      if(!move_arm_warehouse_reader_->getAssociatedMotionPlanRequest("", planning_scene_creation_times[i],
                                                                     motion_plan_request_stage_names[0],
                                                                     motion_plan_request,ID)) {
        ROS_WARN_STREAM("Specific motion plan request bad for stage names " << motion_plan_request_stage_names[0]);
        continue;
      }

      if(motion_plan_request.group_name.empty()) {
        ROS_WARN_STREAM("No group name for motion plan request stage name " << motion_plan_request_stage_names[0]);
        continue;
      }

      bool has_path_constraints = false;
      if(!motion_plan_request.path_constraints.orientation_constraints.empty() ||
         !motion_plan_request.path_constraints.position_constraints.empty()) {
        has_path_constraints = true;
        with_path_constraints_.invocations_++;
      } else {
        without_path_constraints_.invocations_++;
      }

      //just gonna take the first one to get the group name, as it shouldn't matter for this
      cps_->setupForGroupQueries(motion_plan_request.group_name,
                                 motion_plan_request.start_state,
                                 current_link_names_,
                                 current_attached_body_names_);

      std::vector<std::string> joint_trajectory_sources;
      if(!move_arm_warehouse_reader_->getAssociatedJointTrajectorySources("", planning_scene_creation_times[i],
                                                                          joint_trajectory_sources)) {
        ROS_WARN_STREAM("No trajectories associated with creation time " << planning_scene_creation_times[i]);
        continue;
      }
      

      for(unsigned int j = 0; j < joint_trajectory_sources.size(); j++) {
        trajectory_msgs::JointTrajectory jt;
        ros::Duration processing_time;
        if(!move_arm_warehouse_reader_->getAssociatedJointTrajectory("", planning_scene_creation_times[i],
                                                                     joint_trajectory_sources[j], 0,
                                                                     processing_time, 
                                                                     jt)) {
          ROS_WARN_STREAM("Specific joint trajectory not available for " << joint_trajectory_sources[j]);
          continue;
        }
        TrajectoryMetrics tm = calculateTrajectoryMetrics(motion_plan_request.start_state, jt);
        tm.processing_time_ = processing_time;
        overall_metrics_.move_arm_metric_vector_[i].stage_metrics_[joint_trajectory_sources[j]] = tm;
        overall_metrics_.trajectory_source_metrics_[joint_trajectory_sources[j]].addTrajectoryMetric(tm);
        if(has_path_constraints) {
          with_path_constraints_.trajectory_source_metrics_[joint_trajectory_sources[j]].addTrajectoryMetric(tm);
        } else {
          without_path_constraints_.trajectory_source_metrics_[joint_trajectory_sources[j]].addTrajectoryMetric(tm);
        }
      }
      if(has_path_constraints) {
        with_path_constraints_.move_arm_metric_vector_.push_back(overall_metrics_.move_arm_metric_vector_[i]);
      } else {
        without_path_constraints_.move_arm_metric_vector_.push_back(overall_metrics_.move_arm_metric_vector_[i]);
      }
    }
    ROS_INFO_STREAM("Overall metrics");
    printOverallMetrics(overall_metrics_);

    ROS_INFO_STREAM("With path constraints metrics");
    printOverallMetrics(with_path_constraints_);

    ROS_INFO_STREAM("Without path constraints metrics");
    printOverallMetrics(without_path_constraints_);
  }

  void printOverallMetrics(const OverallMetrics& metrics)
  {
    ROS_INFO_STREAM("\n\nMove arm had " << metrics.successes_ << " out of " 
                    << metrics.invocations_ << " successful invocations.  Recoverable errors: " << metrics.recoverable_errors_);

    for(std::map<std::string, unsigned int>::const_iterator it = metrics.failure_stage_.begin();
        it != metrics.failure_stage_.end();
        it++) {
      ROS_INFO_STREAM("Failure stage " << it->first << " instances " << it->second);
    }

    for(std::map<std::string, TrajectorySourceMetrics>::const_iterator it = metrics.trajectory_source_metrics_.begin();
        it != metrics.trajectory_source_metrics_.end();
        it++) {
      ROS_INFO_STREAM("\n");
      ROS_INFO_STREAM("Source: " << it->first);
      ROS_INFO_STREAM("-----------------------");
      ROS_INFO_STREAM("Average average closest collision distance: " 
                      << it->second.total_average_closest_collision_distance_/it->second.getSizeAsDouble());
      ROS_INFO_STREAM("Average overall closest collision distance: " 
                      << it->second.total_overall_closest_collision_distance_/it->second.getSizeAsDouble());
      ROS_INFO_STREAM("Average total path length: " 
                      << it->second.total_total_path_length_/it->second.getSizeAsDouble());
      ROS_INFO_STREAM("Average spline duration: " 
                      << it->second.total_spline_time_/it->second.getSizeAsDouble());
      ROS_INFO_STREAM("Average processing time: " 
                      << it->second.total_processing_time_.toSec()/it->second.getSizeAsDouble());
      ROS_INFO_STREAM("\n");
    }
  }

public:

  ros::Publisher vis_marker_publisher_;
  ros::Publisher vis_marker_array_publisher_;

  std::vector<std::string> current_link_names_;
  std::vector<std::string> current_attached_body_names_;

  move_arm_warehouse::MoveArmWarehouseLoggerReader* move_arm_warehouse_reader_;
  collision_proximity::CollisionProximitySpace* cps_;
  ros::NodeHandle root_nh_;
  OverallMetrics overall_metrics_;
  OverallMetrics with_path_constraints_;
  OverallMetrics without_path_constraints_;
  std::map<std::string, arm_navigation_msgs::JointLimits> joint_limits_map_;
};

int main(int argc, char** argv)
{
  ros::init(argc, argv, "warehouse_metrics_evaluator");

  ros::AsyncSpinner spinner(1); 
  spinner.start();
 
  WarehouseMetricsEvaluator wme;
  wme.calculateMetrics();
  ros::shutdown();
  return 0;
}
