/*
 * Copyright (c) 2011, Willow Garage, Inc.
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the <ORGANIZATION> nor the names of its
 *       contributors may be used to endorse or promote products derived from
 *       this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

// Author: E. Gil Jones

#include <ros/ros.h>

#include <planning_environment/models/collision_models.h>
#include <kinematics_base/kinematics_base.h>
#include <pluginlib/class_loader.h>
 
static const unsigned int TRIES_MAX=10;

inline double gen_rand(double min, double max)
{
  int rand_num = rand()%100+1;
  double result = min + (double)((max-min)*rand_num)/101.0;
  return result;
}

inline geometry_msgs::Quaternion generateRandomUnitQuaternion() {

  geometry_msgs::Quaternion quat;
  quat.x = gen_rand(-1.0, 1.0);
  quat.w = gen_rand(-1.0, 1.0);
  quat.z = gen_rand(-1.0, 1.0);
  quat.w = gen_rand(-1.0, 1.0);

  double mag = sqrt(pow(quat.x, 2.0)+pow(quat.y, 2.0)+pow(quat.z, 2.0)+pow(quat.w, 2.0));

  quat.x /= mag;
  quat.y /= mag;
  quat.z /= mag;
  quat.w /= mag; 

  return quat;
}

class IkPluginBakeoff {
  
public:

  IkPluginBakeoff() : kl_base_("kinematics_base","kinematics::KinematicsBase"),
                      priv_nh_("~")
  {
    priv_nh_.param<std::string>("group", group_, std::string());    
    priv_nh_.param<std::string>("gold_solver_name", gold_solver_name_, std::string());    
    priv_nh_.param<std::string>("comp_solver_name", comp_solver_name_, std::string());    
    priv_nh_.param<double>("workspace_x_extents", workspace_x_extents_, 1.0);
    priv_nh_.param<double>("workspace_y_extents", workspace_y_extents_, 1.0);
    priv_nh_.param<double>("workspace_z_extents", workspace_z_extents_, 1.0);

    cm_ = new planning_environment::CollisionModels("robot_description");

    const planning_models::KinematicModel::JointModelGroup* joint_model_group = cm_->getKinematicModel()->getModelGroup(group_);
    if(joint_model_group == NULL) {
      ROS_WARN_STREAM("No joint group " << group_);
      return;
    }
  
    for(unsigned int i = 0; i < joint_model_group->getJointModels().size(); i++) {
      const planning_models::KinematicModel::JointModel* jm = joint_model_group->getJointModels()[i];
      std::pair<double, double> bounds;
      jm->getVariableBounds(jm->getName(), bounds);
      joint_bounds_map_[jm->getName()] = bounds;
    }

    try
    {
      ks_gold_ = kl_base_.createClassInstance(gold_solver_name_);
    }
    catch(pluginlib::PluginlibException& ex)
    {
      ROS_ERROR("The plugin failed to load. Error1: %s", ex.what());    //handle the class failing to load
      return;
    }
    if(!ks_gold_->initialize(group_)) {
      ROS_ERROR("Gold solver couldn't initialize group");
      return;
    }

    try
    {
      ks_comp_ = kl_base_.createClassInstance(comp_solver_name_);
    }
    catch(pluginlib::PluginlibException& ex)
    {
      ROS_ERROR("The plugin failed to load. Error1: %s", ex.what());    //handle the class failing to load
      return;
    }
    if(!ks_comp_->initialize(group_)) {
      ROS_ERROR("Comp solver couldn't initialize group");
      return;
    }
  }

  arm_navigation_msgs::RobotState generateRandomRobotStateWithinLimits() {
    
    arm_navigation_msgs::RobotState rs;
    rs.joint_state.header.frame_id = cm_->getWorldFrameId();
    rs.joint_state.header.stamp = ros::Time::now();

    for(std::map<std::string, std::pair<double, double> >::iterator it = joint_bounds_map_.begin();
        it != joint_bounds_map_.end();
        it++) {
      rs.joint_state.name.push_back(it->first);
      rs.joint_state.position.push_back(gen_rand(it->second.first, it->second.second));
    }
    return rs;
  }

  std::vector<double> generateRandomRobotVectorWithinLimits() {
    
    std::vector<std::string> ik_names = ks_gold_->getJointNames();

    std::vector<double> rv;

    for(unsigned int i = 0; i < ik_names.size(); i++) {
      std::pair<double, double>& bounds = joint_bounds_map_[ik_names[i]];
      rv.push_back(gen_rand(bounds.first, bounds.second));
    }
    return rv;
  }


  geometry_msgs::Pose generateRandomPoseInWorkspace() {
    
    geometry_msgs::Pose rp;
    rp.position.x = gen_rand(-workspace_x_extents_, workspace_x_extents_);
    rp.position.y = gen_rand(-workspace_y_extents_, workspace_y_extents_);
    rp.position.z = gen_rand(-workspace_z_extents_, workspace_z_extents_);

    rp.orientation = generateRandomUnitQuaternion();

    return rp;
    
  }

  void runBakeoff(unsigned int num) {

    unsigned int tot_comp_failures = 0;
    unsigned int tot_comp_tries = 0;

    double gold_total_good_time;
    double gold_total_bad_time;

    double comp_total_good_time;
    double comp_total_bad_time;

    unsigned int successful_comp_count = 0;

    unsigned int tot_gold_bad_count = 0;

    std::vector<double> solution_gold;
    for(unsigned int i = 0; i < 1; i++) {
      
      geometry_msgs::Pose good_pose;
      while(1) {
        int error;
        good_pose = generateRandomPoseInWorkspace();
        ros::WallTime wd = ros::WallTime::now();
        if(!ks_gold_->getPositionIK(good_pose,
                                    generateRandomRobotVectorWithinLimits(),
                                    solution_gold, 
                                    error)) {
          gold_total_bad_time += (ros::WallTime::now()-wd).toSec();
          tot_gold_bad_count++;
          if(ks_gold_->getPositionIK(good_pose,
                                     generateRandomRobotVectorWithinLimits(),
                                     solution_gold, 
                                     error)) {
            ROS_WARN_STREAM("Gold inconsistent");
          }
        } else {
          gold_total_good_time += (ros::WallTime::now()-wd).toSec();
          break;
        }
      }

      unsigned int comp_count = 1;
      while(comp_count != TRIES_MAX) {
        std::vector<double> solution_comp;
        int error;
        ros::WallTime wd = ros::WallTime::now();        
        if(ks_comp_->getPositionIK(good_pose,
                                   generateRandomRobotVectorWithinLimits(),
                                   solution_comp, 
                                   error)) {
          comp_total_good_time += (ros::WallTime::now()-wd).toSec();
          //TODO - check error with fk
          break;
        }
        comp_total_bad_time += (ros::WallTime::now()-wd).toSec();
        comp_count++;
      }
      if(comp_count == TRIES_MAX) {
        tot_comp_failures++;
      } else {
        successful_comp_count  += comp_count;
      } 
      tot_comp_tries += comp_count;
    }

    ROS_INFO_STREAM("Average samples to produce good gold sample is " << (tot_gold_bad_count*1.0)/(num*1.0));
    ROS_INFO_STREAM("Average gold bad time " << gold_total_bad_time/(tot_gold_bad_count*1.0));
    ROS_INFO_STREAM("Average gold good time " << gold_total_good_time/(num*1.0));
    ROS_INFO_STREAM("Average comp bad time " << comp_total_bad_time/(tot_comp_tries*1.0));
    ROS_INFO_STREAM("Average comp good time " << comp_total_good_time/(num-tot_comp_failures*1.0));
                    
    ROS_INFO_STREAM("Average comp successful tries " << (successful_comp_count*1.0)/(num-tot_comp_failures*1.0));

    ROS_INFO_STREAM("Total failures is " << tot_comp_failures << " out of " << num);
  }

private:

  pluginlib::ClassLoader<kinematics::KinematicsBase> kl_base_;

  kinematics::KinematicsBase* ks_gold_;
  kinematics::KinematicsBase* ks_comp_;

  std::map<std::string, std::pair<double, double> > joint_bounds_map_;
  planning_environment::CollisionModels* cm_;

  ros::NodeHandle priv_nh_;
  std::string group_;
  std::string gold_solver_name_;
  std::string comp_solver_name_;

  double workspace_x_extents_;
  double workspace_y_extents_;
  double workspace_z_extents_;

};

int main(int argc, char** argv)
{
  srand(time(0));

  ros::init(argc, argv, "ik_plugin_bakeoff");

  IkPluginBakeoff bake;

  bake.runBakeoff(1000);

  ros::shutdown();
}

